{% extends "layout.html" %}
{% block body %}
{{ super() }}
<section id="home" class="home">
    <div class="container" style="text-align: center">
        <h1 class="logo" style="text-align: center">Method not allowed (405)</h1>
        <p style="text-align: center" class="fst-italic">The method of your request is not allowed.</p>
    </div>
</section>
{% endblock %}
