# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: nebula.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0cnebula.proto\x12\x06nebula"\xe4\x02\n\x07Wrapper\x12\x0e\n\x06source\x18\x01 \x01(\t\x12\x35\n\x11\x64iscovery_message\x18\x02 \x01(\x0b\x32\x18.nebula.DiscoveryMessageH\x00\x12\x31\n\x0f\x63ontrol_message\x18\x03 \x01(\x0b\x32\x16.nebula.ControlMessageH\x00\x12\x37\n\x12\x66\x65\x64\x65ration_message\x18\x04 \x01(\x0b\x32\x19.nebula.FederationMessageH\x00\x12-\n\rmodel_message\x18\x05 \x01(\x0b\x32\x14.nebula.ModelMessageH\x00\x12\x37\n\x12\x63onnection_message\x18\x06 \x01(\x0b\x32\x19.nebula.ConnectionMessageH\x00\x12\x33\n\x10response_message\x18\x07 \x01(\x0b\x32\x17.nebula.ResponseMessageH\x00\x42\t\n\x07message"\x9e\x01\n\x10\x44iscoveryMessage\x12/\n\x06\x61\x63tion\x18\x01 \x01(\x0e\x32\x1f.nebula.DiscoveryMessage.Action\x12\x10\n\x08latitude\x18\x02 \x01(\x02\x12\x11\n\tlongitude\x18\x03 \x01(\x02"4\n\x06\x41\x63tion\x12\x0c\n\x08\x44ISCOVER\x10\x00\x12\x0c\n\x08REGISTER\x10\x01\x12\x0e\n\nDEREGISTER\x10\x02"\x9a\x01\n\x0e\x43ontrolMessage\x12-\n\x06\x61\x63tion\x18\x01 \x01(\x0e\x32\x1d.nebula.ControlMessage.Action\x12\x0b\n\x03log\x18\x02 \x01(\t"L\n\x06\x41\x63tion\x12\t\n\x05\x41LIVE\x10\x00\x12\x0c\n\x08OVERHEAD\x10\x01\x12\x0c\n\x08MOBILITY\x10\x02\x12\x0c\n\x08RECOVERY\x10\x03\x12\r\n\tWEAK_LINK\x10\x04"\xcd\x01\n\x11\x46\x65\x64\x65rationMessage\x12\x30\n\x06\x61\x63tion\x18\x01 \x01(\x0e\x32 .nebula.FederationMessage.Action\x12\x11\n\targuments\x18\x02 \x03(\t\x12\r\n\x05round\x18\x03 \x01(\x05"d\n\x06\x41\x63tion\x12\x14\n\x10\x46\x45\x44\x45RATION_START\x10\x00\x12\x0e\n\nREPUTATION\x10\x01\x12\x1e\n\x1a\x46\x45\x44\x45RATION_MODELS_INCLUDED\x10\x02\x12\x14\n\x10\x46\x45\x44\x45RATION_READY\x10\x03"A\n\x0cModelMessage\x12\x12\n\nparameters\x18\x01 \x01(\x0c\x12\x0e\n\x06weight\x18\x02 \x01(\x03\x12\r\n\x05round\x18\x03 \x01(\x05"l\n\x11\x43onnectionMessage\x12\x30\n\x06\x61\x63tion\x18\x01 \x01(\x0e\x32 .nebula.ConnectionMessage.Action"%\n\x06\x41\x63tion\x12\x0b\n\x07\x43ONNECT\x10\x00\x12\x0e\n\nDISCONNECT\x10\x01"#\n\x0fResponseMessage\x12\x10\n\x08response\x18\x01 \x01(\tb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "nebula_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _globals["_WRAPPER"]._serialized_start = 25
    _globals["_WRAPPER"]._serialized_end = 381
    _globals["_DISCOVERYMESSAGE"]._serialized_start = 384
    _globals["_DISCOVERYMESSAGE"]._serialized_end = 542
    _globals["_DISCOVERYMESSAGE_ACTION"]._serialized_start = 490
    _globals["_DISCOVERYMESSAGE_ACTION"]._serialized_end = 542
    _globals["_CONTROLMESSAGE"]._serialized_start = 545
    _globals["_CONTROLMESSAGE"]._serialized_end = 699
    _globals["_CONTROLMESSAGE_ACTION"]._serialized_start = 623
    _globals["_CONTROLMESSAGE_ACTION"]._serialized_end = 699
    _globals["_FEDERATIONMESSAGE"]._serialized_start = 702
    _globals["_FEDERATIONMESSAGE"]._serialized_end = 907
    _globals["_FEDERATIONMESSAGE_ACTION"]._serialized_start = 807
    _globals["_FEDERATIONMESSAGE_ACTION"]._serialized_end = 907
    _globals["_MODELMESSAGE"]._serialized_start = 909
    _globals["_MODELMESSAGE"]._serialized_end = 974
    _globals["_CONNECTIONMESSAGE"]._serialized_start = 976
    _globals["_CONNECTIONMESSAGE"]._serialized_end = 1084
    _globals["_CONNECTIONMESSAGE_ACTION"]._serialized_start = 1047
    _globals["_CONNECTIONMESSAGE_ACTION"]._serialized_end = 1084
    _globals["_RESPONSEMESSAGE"]._serialized_start = 1086
    _globals["_RESPONSEMESSAGE"]._serialized_end = 1121
# @@protoc_insertion_point(module_scope)
