html, body {
    height: 100%;
    margin: 0;
}

body {
    font-family: "Inter", sans-serif;
    color: #444444;
    display: flex;
    flex-direction: column;
}

#main {
    flex: 1;
}

a {
    color: #ff9b7f;
    text-decoration: none;
}

a:hover {
    color: #ff6940;
    text-decoration: none;
}

h1,
h2,
h3,
h4,
h5,
h6 {
    font-family: "Inter", sans-serif;
}

#preloader {
    position: fixed;
    top: 0;
    left: 0;
    right: 0;
    bottom: 0;
    z-index: 9999;
    overflow: hidden;
    background: #fff;
}

#preloader:before {
    content: "";
    position: fixed;
    top: calc(50% - 30px);
    left: calc(50% - 30px);
    border: 6px solid #ff4a17;
    border-top-color: #ffe9e3;
    border-radius: 50%;
    width: 60px;
    height: 60px;
    animation: animate-preloader 1s linear infinite;
}

@keyframes animate-preloader {
    0% {
        transform: rotate(0deg);
    }

    100% {
        transform: rotate(360deg);
    }
}

#header {
    transition: all 0.5s;
    z-index: 997;
    padding: 10px 0;
    background: #1d2253;
}

#header.header-scrolled,
#header.header-inner-pages {
    background: #1d2253;
    padding: 10px 0;
}

#header .logo {
    font-size: 32px;
    margin: 0;
    padding: 0;
    line-height: 1;
    font-weight: 700;
    letter-spacing: 2px;
}

#header .logo a {
    color: #fff;
}

#header .logo img {
    max-height: 40px;
}

.navbar {
    padding: 0;
    z-index: 150;
}

.navbar ul {
    margin: 0;
    padding: 0;
    display: flex;
    list-style: none;
    align-items: center;
}

.navbar li {
    position: relative;
}

.navbar>ul>li {
    white-space: nowrap;
    padding: 8px 12px;
}

.navbar a,
.navbar a:focus {
    display: flex;
    align-items: center;
    position: relative;
    padding: 0 3px;
    font-size: 14px;
    text-transform: uppercase;
    font-weight: 600;
    color: rgba(255, 255, 255, 0.9);
    white-space: nowrap;
    transition: 0.3s;
}

.navbar a i,
.navbar a:focus i {
    font-size: 12px;
    line-height: 0;
    margin-left: 5px;
}

.navbar>ul>li>a:before {
    content: "";
    position: absolute;
    width: 100%;
    height: 2px;
    bottom: -6px;
    left: 0;
    background-color: #ff4a17;
    visibility: hidden;
    width: 0px;
    transition: all 0.3s ease-in-out 0s;
}

.navbar a:hover:before,
.navbar li:hover>a:before,
.navbar .active:before {
    visibility: visible;
    width: 100%;
}

.navbar a:hover,
.navbar .active,
.navbar .active:focus,
.navbar li:hover>a {
    color: #fff;
}

.navbar .getstarted,
.navbar .getstarted:focus {
    padding: 8px 25px;
    margin-left: 30px;
    border-radius: 4px;
    color: #fff;
    border: 2px solid #fff;
}

.navbar .getstarted:hover,
.navbar .getstarted:focus:hover {
    color: #fff;
    background: #fd3800;
    border-color: #ff4a17;
}

.navbar>ul>li>.getstarted:before {
    visibility: hidden;
}

.navbar .dropdown ul {
    display: block;
    position: absolute;
    left: 14px;
    top: calc(100% + 30px);
    margin: 0;
    padding: 10px 0;
    z-index: 99;
    opacity: 0;
    visibility: hidden;
    background: #fff;
    box-shadow: 0px 0px 30px rgba(127, 137, 161, 0.25);
    transition: 0.3s;
    border-radius: 4px;
}

.navbar .dropdown ul li {
    min-width: 200px;
}

.navbar .dropdown ul a {
    padding: 10px 20px;
    font-size: 14px;
    text-transform: none;
    color: #15222b;
    font-weight: 400;
}

.navbar .dropdown ul a i {
    font-size: 12px;
}

.navbar .dropdown ul a:hover,
.navbar .dropdown ul .active:hover,
.navbar .dropdown ul li:hover>a {
    color: #ff4a17;
}

.navbar .dropdown:hover>ul {
    opacity: 1;
    top: 100%;
    visibility: visible;
}

.navbar .dropdown .dropdown ul {
    top: 0;
    left: calc(100% - 30px);
    visibility: hidden;
}

.navbar .dropdown .dropdown:hover>ul {
    opacity: 1;
    top: 0;
    left: 100%;
    visibility: visible;
}

@media (max-width: 1366px) {
    .navbar .dropdown .dropdown ul {
        left: -90%;
    }

    .navbar .dropdown .dropdown:hover>ul {
        left: -100%;
    }
}

.mobile-nav-toggle {
    color: #fff;
    font-size: 28px;
    cursor: pointer;
    display: none;
    line-height: 0;
    transition: 0.5s;
}

@media (max-width: 991px) {
    .mobile-nav-toggle {
        display: block;
    }

    .navbar ul {
        display: none;
    }
}

.navbar-mobile {
    position: fixed;
    overflow: hidden;
    top: 0;
    right: 0;
    left: 0;
    bottom: 0;
    background: rgba(4, 7, 9, 0.9);
    transition: 0.3s;
    z-index: 999;
}

.navbar-mobile .mobile-nav-toggle {
    position: absolute;
    top: 15px;
    right: 15px;
}

.navbar-mobile ul {
    display: block;
    position: absolute;
    top: 55px;
    right: 15px;
    bottom: 15px;
    left: 15px;
    padding: 10px 0;
    border-radius: 10px;
    background-color: #fff;
    overflow-y: auto;
    transition: 0.3s;
}

.navbar-mobile a,
.navbar-mobile a:focus {
    padding: 10px 20px;
    font-size: 15px;
    color: #15222b;
}

.navbar-mobile a:hover,
.navbar-mobile .active,
.navbar-mobile li:hover>a {
    color: #ff4a17;
}

.navbar-mobile .getstarted,
.navbar-mobile .getstarted:focus {
    margin: 15px;
    color: #15222b;
    border: 2px solid #15222b;
}

.navbar-mobile .dropdown ul {
    position: static;
    display: none;
    margin: 10px 20px;
    padding: 10px 0;
    z-index: 99;
    opacity: 1;
    visibility: visible;
    background: #fff;
    box-shadow: 0px 0px 30px rgba(127, 137, 161, 0.25);
}

.navbar-mobile .dropdown ul li {
    min-width: 200px;
}

.navbar-mobile .dropdown ul a {
    padding: 10px 20px;
}

.navbar-mobile .dropdown ul a i {
    font-size: 12px;
}

.navbar-mobile .dropdown ul a:hover,
.navbar-mobile .dropdown ul .active:hover,
.navbar-mobile .dropdown ul li:hover>a {
    color: #ff4a17;
}

.navbar-mobile .dropdown>.dropdown-active {
    display: block;
}

.no-padding {
    padding-left: 0 !important;
    padding-right: 0 !important;
}

.nav-item label {
    cursor: pointer;
    margin-bottom: 0;
}

.navbar-brand {
    font-size: 25px;
}

.navbar-form {
    margin-right: 0;
    margin-left: 0;
}

section {
    padding: 20px 0;
    overflow: hidden;
}

.section-bg {
    background-color: #f0f4f8;
}

.alerts {
    z-index: 100;
    position: absolute;
    text-align: center;
}

.container {
    max-width: 90%;
}

.colaboration {
    background: #f0f4f8;
    padding: 15px 0;
    text-align: center;
}

.colaboration img {
    max-width: 100%;
    transition: all 0.4s ease-in-out;
    display: inline-block;
    padding: 15px 0;
    filter: grayscale(100);
}

.colaboration img:hover {
    filter: none;
    transform: scale(1.15);
}

@media (max-width: 768px) {
    .colaboration img {
        max-width: 80%;
    }
}

.base .container {
    box-shadow: 0px 2px 15px rgba(0, 0, 0, 0.1);
    padding-bottom: 15px;
}

.base .content {
    font-size: 15px;
}

.base .content h3 {
    font-weight: 700;
    font-size: 24px;
    color: #263d4d;
}

.base .content ul {
    list-style: none;
    padding: 0;
}

.base .content ul li {
    padding-bottom: 10px;
    padding-left: 28px;
    position: relative;
}

.base .content ul i {
    font-size: 24px;
    color: #ff4a17;
    position: absolute;
    left: 0;
    top: -2px;
}

.base .content p:last-child {
    margin-bottom: 0;
}

.about .container {
    box-shadow: 0px 2px 15px rgba(0, 0, 0, 0.1);
    padding-bottom: 15px;
}

.about .count-box {
    padding: 60px 0;
    width: 100%;
}

.about .count-box i {
    display: block;
    font-size: 48px;
    color: #a1bdd1;
    float: left;
    line-height: 0;
}

.about .count-box span {
    font-size: 28px;
    line-height: 25px;
    display: block;
    font-weight: 700;
    color: #365870;
    margin-left: 60px;
}

.about .count-box p {
    padding: 5px 0 0 0;
    margin: 0 0 0 60px;
    font-family: "Inter", sans-serif;
    font-weight: 600;
    font-size: 14px;
    color: #2e4b5e;
}

.about .count-box a {
    font-weight: 600;
    display: block;
    margin-top: 20px;
    color: #2e4b5e;
    font-size: 15px;
    font-family: "Inter", sans-serif;
    transition: ease-in-out 0.3s;
}

.about .count-box a:hover {
    color: #477392;
}

.about .content {
    font-size: 15px;
}

.about .content h3 {
    font-weight: 700;
    font-size: 24px;
    color: #263d4d;
}

.about .content ul {
    list-style: none;
    padding: 0;
}

.about .content ul li {
    padding-bottom: 10px;
    padding-left: 28px;
    position: relative;
}

.about .content ul i {
    font-size: 24px;
    color: #ff4a17;
    position: absolute;
    left: 0;
    top: -2px;
}

.about .content p:last-child {
    margin-bottom: 0;
}

.alert {
    margin: 5px 5px 5px 5px;
}

@media (min-width: 768px) {
    .nav>li>a {
        padding: 5px 10px;
    }

    .navbar-nav>li>a {
        padding-top: 15px;
        padding-bottom: 15px;
    }
}

#footer {
    background: #1d2253;
    padding: 0 0 10px 0;
    color: #fff;
    font-size: 14px;
}

#footer .copyright {
    text-align: center;
    padding-top: 10px;
}

div {
    text-align: justify;
}

h1,
h2,
h3,
h4,
h5,
h6 {
    text-align: left;
    font-family: "Inter", sans-serif;
}

hr.styled {
    border: 0;
    height: 1px;
    background-image: linear-gradient(to right, rgba(0, 0, 0, 0), rgba(0, 0, 0, 0.75), rgba(0, 0, 0, 0));
}

.td-icons .label {
    font-size: 12px;
    line-height: 0.8 !important;
}

.table {
    display: block !important;
}

.table-centered td,
.table-centered th {
    text-align: center;
    vertical-align: middle;
}

.table-centered form {
    display: inline-block;
}

.table-responsive>.table>thead>tr>th,
.table-responsive>.table>tbody>tr>th,
.table-responsive>.table>tfoot>tr>th,
.table-responsive>.table>thead>tr>td,
.table-responsive>.table>tbody>tr>td,
.table-responsive>.table>tfoot>tr>td {
    white-space: nowrap;
}

@media screen and (max-width: 1530px) {
    .table-responsive {
        width: 100%;
        margin-bottom: 15px;
        overflow-x: auto;
        -ms-overflow-style: -ms-autohiding-scrollbar;
        -webkit-overflow-scrolling: touch;
    }
}

td {
    vertical-align: middle !important;
}

.truncate-text {
    max-width: 150px;
    overflow: hidden;
    text-overflow: ellipsis;
    white-space: nowrap;
    cursor: pointer;
}

.tooltip-inner {
    max-width: 500px !important;
}

.btn-cream {
    color: #000000;
    background-color: #FFEFCB;
    border-color: #FFEFCB;
}

.btn-light {
    color: #ffffff;
    background-color: #4d517d;
    border-color: #4d517d;
}

.btn-light:hover {
    color: #ffffff;
    background-color: #4d517de0;
    border-color: #4d517de0;
}

.btn-dark {
    color: #ffffff;
    background-color: #1d2253;
    border-color: #1d2253;
}

.btn-dark:hover {
    color: #ffffff;
    background-color: #1d2253d2;
    border-color: #1d2253d2;
}

.form-group {
    margin-bottom: 1rem !important;
    vertical-align: baseline !important;
}

@media (min-width: 768px) {

    .form-group.form-group-tiny {
        margin-bottom: 0 !important;
        vertical-align: baseline !important;
    }

}

.dropdown-toggle:empty::after {
    margin-left: 0;
}

.container-shadow {
    padding: 20px;
    background-color: #fff;
    box-shadow: 0px 5px 10px rgba(0, 0, 0, 0.1);
    border-radius: 5px;
    margin: auto auto 20px !important;
}

.container-shadow.tiny {
    padding: 10px;
}

.container-shadow.grey {
    background-color: #f5f5f5;
}

.container-shadow h3 {
    color: #2c3e50;
    font-size: 24px;
}

.container-shadow p {
    color: #7f8c8d;
}

.container-shadow li {
    color: #7f8c8d;
}

#btn-up {
    display: none;
    position: fixed;
    bottom: 20px;
    right: 30px;
    z-index: 99;
    border: none;
    outline: none;
    color: white;
    cursor: pointer;
    border-radius: 50%;
    font-size: 18px;
    transition: display 0.5s;
}

#form-configurations {
    padding: 10px;
}
