class Role:
    """
    This class defines the participant roles of the platform.
    """

    TRAINER = "trainer"
    AGGREGATOR = "aggregator"
    PROXY = "proxy"
    IDLE = "idle"
    SERVER = "server"
