__version__ = "0.0.1"
__description__ = "NEBULA: A Platform for Decentralized Federated Learning"
__long_description__ = "NEBULA: A Platform for Decentralized Federated Learning"
__author__ = "Enrique Tomás Martínez Beltrán"
__long_description_content_type__ = "text/markdown"
__keywords__ = "federated learning, decentralized federated learning, machine learning, deep learning, neural networks, collaborative learning"
