{% extends "layout.html" %}
{% block body %}
<!-- Print statistics_url using jinja2 -->
<!-- {{ statistics_url | safe }} -->
<iframe src="{{ statistics_url | safe }}" width="100%" height="100%" style="border: none; overflow: hidden"></iframe>
<style>
    .navbar {
        margin-bottom: 0;
    }
</style>
{% endblock %}

{% block footer %}
{% endblock %}
